

/*

Essential stuff for XUL-Node

	* error handling
	* global events
	* global utility functions
	* javascript OO support
	* debug support
	* client server communications

*/

// sniffing -------------------------------------------------------------------

var VERSION = {
	"NS6up": navigator.userAgent.indexOf("Gecko")  != -1,
	"IE6up": navigator.userAgent.indexOf("MSIE 6") != -1
};

// global events --------------------------------------------------------------

var Class_LAST_ERROR = {'message': ''};

window.onerror = function ()
	{ if (Class_LAST_ERROR) dumpln(Class_LAST_ERROR.message) }

// error handling -------------------------------------------------------------

function Throw(a, b) {
	var message       = b? (a.message || a.description) + "\n" + b: a;
	var exception     = b? a: new Error;
	exception.message = exception.description = message;
	Class_LAST_ERROR  = exception;
	throw exception;
}

// global utilities -----------------------------------------------------------

function isDefined (value) { return value != undefined    }

function argsAsArray (args) {
	var results = [];
	for (var i in args) results.push(args[i]);
	return results;
}

function trim (s) {
	if (!s) return "";
	return s.replace(/^\s+/, "").replace(/\s+$/, "");
}

function dumpln (message) { dump(message + "\n") }

function use (className) { Class_Use(className) }

// OO support -----------------------------------------------------------------

$ForPrototype = false;
Class_Classes = {};

function Class (className, parentClassName) {
	Class_CheckClassName(className);
	if (parentClassName) {
		Class_CheckClassName(parentClassName);
		use(parentClassName);
	}
	var classObj = eval(
		className + " = function () {" +
			"if (!this.init) Throw('Class not used: " + className + "');" +
			"if ($ForPrototype) return;" +
			"this.init.apply(this, arguments);" +
		"}"
	);
	classObj.className       = className;
	Class_Classes[className] = classObj;
	if (parentClassName) classObj.parentClassName = parentClassName;
	_ = classObj.prototype;
	_.$classObj = classObj;
}

function Class_Singleton () {
	var classObj = _.$classObj;
	var singleton;
	classObj.get = function () {
		if (singleton) return singleton;
		singleton = new classObj;
		return singleton;
	};
}

function Class_Use (className) { try {
	var classObj = Class_Classes[className];
	if (!classObj) Util_Conduit.get().use(className);
	classObj = Class_Classes[className];
	if (!classObj) Throw("Class not declared before use: " + className);
	if (classObj.isUsed) return;

	var proto = classObj.prototype;
	var methodNames = Class_getMethodNames(proto);
	var i; for (i in methodNames) {
		var methodName      = methodNames[i];
		var method          = proto[methodName];
		var wrapped         = Class_wrapMethod(className, methodName, method);
		wrapped.$methodName = methodName;
		proto[methodName]   = wrapped;
	}

	var parentClassName = classObj.parentClassName;
	if (parentClassName) {
		var parentClassObj  = Class_Classes[parentClassName];
		if (!parentClassObj) Throw(
			"Base class not declared before inheritance: "
			+ className + ":" + parentClassName
		);
		if (!parentClassObj.isUsed) Throw(
			"Base class not used before inheritance: "
			+ className + ":" + parentClassName
		);
		$ForPrototype = true;
		var parentInstance = new parentClassObj;
		$ForPrototype = false;
		var i; for (i in methodNames) {
			var methodName   = methodNames[i];
			var parentMethod = parentInstance[methodName];
			var childMethod  = proto[methodName];
			if (parentMethod) {
				parentInstance[parentClassName + "_" + methodName] =
					parentInstance[methodName];
			}
			parentInstance[methodName] = childMethod;
		}
		classObj.prototype = parentInstance;
	}
	proto = classObj.prototype;
	if (!proto.init) proto.init = function () {};
	proto.$className = className;
	proto.$classObj  = classObj;
	classObj.isUsed  = true;
} catch (e) { Throw(e, "Using: " + className) }}

function Class_wrapMethod (className, methodName, method) {
	return !Util_Debug.get().getEnabled()?
		method:
		function () { try {

			Class_PreHandler(this, className, methodName, arguments);
			var returnValue = method.apply(this, arguments);
			Class_PostHandler(this, className, methodName, returnValue);
			return returnValue;

		} catch (_e) { Throw(_e,
			"* " + className + ":" + methodName + "(" + this.$className + ")"
		)}}
}

function Class_getMethodNames (o) {
	var methodNames = [];
	for (methodName in o)
		if (typeof o[methodName]  == "function" && !methodName.match(/^\$/))
			methodNames.push(methodName);
	return methodNames;
}

function Class_asString (o, verbose) {
	var result = "";
	var p; for (p in o) result += p + (verbose? ":" + o[p]: "");
	return result;
}

function Class_CheckClassName (className) {
	if (className.match(/\s/))
		Throw("Not a legal class name: [" + className + "]");
}

function Class_PreHandler (object, className, methodName, arguments) {
	return;
	if (methodName.match(/Min/))
		dumpln(
			className + ":" + methodName + "(" + object.$className + ") : " +
			argsAsArray(arguments).join(",")
		);
}

function Class_PostHandler (object, className, methodName, returnValue) {
	return;
	if (methodName.match(/Min/))
		dumpln(
			className + ":" + methodName + "(" + object.$className + ") : " +
			returnValue
		);
}

// debug support --------------------------------------------------------------

Class("Util_Debug");
Class_Singleton();

_.init       = function ()        { this.setEnabled(false) }
_.getEnabled = function ()        { return this.isEnabled }
_.setEnabled = function (value)   { this.isEnabled = value }
_.dumpBare   = function (message) { dump(message + "\n") }

use("Util_Debug");

// client server communications -----------------------------------------------

Class("Util_Conduit");
Class_Singleton();

_.init = function () {
	var hash = location.hash;
	Util_Debug.get().setEnabled(
		Util_Debug.get().getEnabled()? true:
		hash.substr(1) == 1? true: false
	);
	var pathname   = location.pathname.replace(/\/[^\/]+$/, "");
	var port       = location.port;
	port           = port? ':' + port: '';
	this.serverURL = 'http://'+location.hostname+port+pathname+"/";
	this.xmlHTTP   = (VERSION.IE6up)?
		new ActiveXObject("Microsoft.XMLHTTP"):
		new XMLHttpRequest;
}

_.request = function (request, payloadXML) { try {
	var xmlHTTP = this.xmlHTTP;
	xmlHTTP.overrideMimeType("text/xml");
	xmlHTTP.open("POST", this.serverURL + request, false);

	try { xmlHTTP.send(payloadXML || null) }
		catch (e) { Throw(e, "No response from server") }

	var response = new String(xmlHTTP.responseText);
	if (xmlHTTP.status != "200") Throw(
		"Cannot send request: " + xmlHTTP.statusText + "\n" +
		xmlHTTP.responseText
	);
	return response;
} catch (e) { Throw(e,
	"While requesting url: " + this.serverURL + ", request: " + request
)}}

_.use = function (className) { try {
	if (VERSION.NS6up)
		Throw("Cannot load classes in NS6up [" + className + "]");
	window.status = "Loading " + className + "...";
	var fileName = className.replace(
		/^Widget_/,
		"Widget_" + (VERSION.NS6up? "NS6": "IE6") + "_"
	);
	var libDir = "jslib/";
	var classText;
	try { classText = this.request(
		libDir + fileName.replace(/_/g, "/") + ".js"
	) } catch (e) { Throw(e, "Requesting class from server") }
	try { eval(classText.toString()) }
		catch(e) { Throw(e, "Evaling class text") }
	window.status = "Loaded " + className;
} catch (e) { Throw(e, "Importing: " + className) }}

use("Util_Conduit");
Util_Conduit.get();



Class("Client_ServerResponse");

Client_ServerResponse.wordSeperator = String.fromCharCode(1);
Client_ServerResponse.lineSeperator = String.fromCharCode(2);

_.init = function (response) {
	this.message  = response;
	this.commands = this.parseCommands();
	if (response.match(/^ERROR/)) Throw("Server side error. " + response);
//	this.dumpResponse();
}

_.parseCommands = function () {
	var outLines = [];
	if (this.message == 'null') return outLines;
	var inLines = this.message.split(this.$classObj.lineSeperator);
	var inLine;
	for (inLine in inLines) {
		inLine = inLines[inLine];
		if (!inLine.match(/\w/)) continue;
		var params  = inLine.split(this.$classObj.wordSeperator);
		outLines.push({
			'nodeId'    : params[0],
			'methodName': params[1],
			'arg1'      : params[2],
			'arg2'      : params[3],
			'arg3'      : params[4]
		});
	}
	return outLines;
}

_.dumpResponse = function () {
	var commands = this.getCommands();
	dumpln("* received response (" + commands.length + " lines):");
	var command;
	for (command in commands) {
		command = commands[command];
		dumpln(
			"   " + this.pad(command['nodeId'], 4) + '.' + command['methodName'] +
			'(' + command['arg1'] + ', ' + command['arg2'] + ")"
		);
	}
}

_.pad = function (input, length) {
	var inLength = input.length;
	if (inLength >= length) return input;
	padLength = length - inLength;
	var count;
	for (count = 0; count < padLength; count++) input = ' ' + input;
	return input;
}

_.getCommands = function () { return this.commands }
_.getMessage  = function () { return this.message  }




use("Client_ServerResponse");

Class("Client_ServerProxy");

_.init = function () {
	this.xml = new DOMParser();
	this.requestCount    = 0;
	this.applicationName = location.search.substr(1) || false;
	this.requestPrefix   = "xul";
	this.sessionId       = null;
}

_.boot =  function (applicationName) {
	this.applicationName = applicationName || this.applicationName;
	var event = {"type": "boot"};
	if (this.applicationName)
		event.name = this.applicationName;
	return this.request(event, 1);
}

_.event = function (event) {
	if (!this.sessionId) Throw("firing event with no session");
	event.type    = "event";
	event.session = this.sessionId;
	return this.request(event);
}

_.request = function (event, isBoot) {
	event.requestCount = ++this.requestCount;
	var payloadXML     = this.getPayloadAsXML(event);
	var rawResponse    = Util_Conduit.get().
		request(this.requestPrefix, payloadXML);
	if (isBoot) {
		this.sessionId = (rawResponse.split("\n"))[0];
		rawResponse    = rawResponse.replace(/.*\n/, '');
	}
	return new Client_ServerResponse(rawResponse);
}

_.getPayloadAsXML = function (event) {
	var root = this.xml.parseFromString
		('<?xml version="1.0" encoding="UTF-8"?><xul></xul>', "text/xml");
	var doc = root.documentElement;
	var key, value;
	for (key in event) {
		var element  = root.createElement(key);
		element.appendChild(root.createTextNode(event[key]));
		doc.appendChild(element);
	}
	return root;
}



Class("Client_Runner");
Class_Singleton();

_.init = function () { this.document = window.document }

_.run = function (response) {
	this.resetBuffers();

	var commands = response.getCommands();
	var command;
	for (command in commands)
		this.runCommand(commands[command]);

	var roots = this.newNodeRoots;
	var parentId;
	for (parentId in roots)
		this.addElementAtIndex(this.getNode(parentId), roots[parentId]);

	var lateCommands = this.lateCommands;
	for (command in lateCommands) {
		command = lateCommands[command];
		this.commandSetNode
			(command['nodeId'], command['arg1'], command['arg2']);
	}
}

// commands -------------------------------------------------------------------

_.runCommand = function (command) {
	var nodeId     = command['nodeId'];
	var methodName = command['methodName'];
	var arg1       = command['arg1'];
	var arg2       = command['arg2'];
	var arg3       = command['arg3'];
	if (methodName == 'new')
		if (arg1 == 'window')
			this.commandNewWindow(nodeId);
		else
			this.commandNewElement(nodeId, arg1, arg2, arg3);
	else
		if (methodName == 'bye')
			this.commandByeElement(nodeId);
		else
			if (Client_Runner.lateAttributes[arg1])
				this.lateCommands.push(command);
			else
				this.commandSetNode(nodeId, arg1, arg2);
}

_.commandNewWindow = function (nodeId) {
	this.windowId = nodeId;
}

_.commandNewElement = function (nodeId, tagName, parentId, index) { try {
	var element = this.createElement(tagName, nodeId);
	element.setAttribute('_addAtIndex', index);
	this.newNodes[nodeId] = element;

	var parent = this.newNodes[parentId];
	if (parent)
		this.addElementAtIndex(parent, element);
	else
		this.newNodeRoots[parentId] = element;

	// onselect does not bubble
	if (tagName == 'listbox')
		element.setAttribute('onselect', 'window.onselect(event)');
	else if (tagName == 'colorpicker')
		element.setAttribute(
			'onselect',
			'Client_Application.get().fireEvent_Pick({"targetId":"' +
				element.id + '"})'
		);
} catch (e) {
	Throw(e,
		'Cannot create new node: [' + nodeId +
		', ' + tagName + ', ' + parentId + ']'
	);
}}

_.commandSetNode = function (nodeId, key, value) { try {
	var element = this.newNodes[nodeId];

	if (!element) element = this.getNode(nodeId);

	if (key == 'textNode') {
		element.appendChild(this.document.createTextNode(value));
		return;
	}
	if (Client_Runner.boleanAttributes[key]) {
		value = (value == 0 || value == '' || value == null)? false: true;
		if (!value)
			element.removeAttribute(key);
		else
			element.setAttribute(key, 'true');
		return;
	}
	if (Client_Runner.simpleMethodAttributes[key]) {
		if (element.tagName == 'window')
			window[key].apply(window, [value]);
		else
			element[key].apply(element, [value]);
		return;
	}
	if (Client_Runner.propertyAttributes[key]) {
		if (key == 'selectedIndex')
			element.setAttribute("suppressonselect", true);
		element[key] = value;
		if (key == 'selectedIndex')
			element.setAttribute("suppressonselect", false);
		return;
	}
	element.setAttribute(key, value);
} catch (e) {
	Throw(e,
		'Cannot do set on node: [' + nodeId + ', ' + key + ', ' + value + ']'
	);
}}

_.commandByeElement = function (nodeId) {
	var node = this.getNode(nodeId);
	node.parentNode.removeChild(node);
}

// private --------------------------------------------------------------------

_.getNode = function (nodeId) {
	var node = this._getNode(nodeId);
	if (!node) Throw("cannot find node by Id: " + nodeId);
	return node;
}

_._getNode = function (nodeId) {
	var node = this.windowId == nodeId?
		this.document.firstChild:
		this.document.getElementById(nodeId);
	return node;
}	

_.createElement = function (tagName, nodeId) {
	var element = tagName.match(/^html_/)?
		this.document.createElementNS(
			'http://www.w3.org/1999/xhtml',
			tagName.replace(/^html_/, '')
		):
		this.document.createElementNS(
			'http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul',
			tagName
		);
	element.id = nodeId;
	return element;
}

_.addElementAtIndex = function (parent, child) {
	var index = child.getAttribute('_addAtIndex');
	child.removeAttribute('_addAtIndex');
	
	if (index == null) {
		parent.appendChild(child);
		return;
	}
	var children = parent.childNodes;
	var count    = children.length;
	if (count == 0)
		parent.appendChild(child);
	else
		parent.insertBefore(child, index == count? null: children[index]);
}

_.resetBuffers = function () {
	this.newNodeRoots = []; // top level parent nodes of those not yet added
	this.newNodes     = []; // nodes not yet added to document
	this.lateCommands = []; // commands to run at latest possible time
}

Client_Runner.boleanAttributes = {
	'disabled'     : true,
	'multiline'    : true,
	'readonly'     : true,
	'checked'      : true,
	'hidden'       : true,
	'default'      : true,
	'grippyhidden' : true
};
Client_Runner.propertyAttributes = {
	'selectedIndex': true
};
Client_Runner.lateAttributes = {
	'selectedIndex': true,
	'sizeToContent': true
};
Client_Runner.simpleMethodAttributes = {
	'sizeToContent'       : true,
	'ensureIndexIsVisible': true
};

 


use("Client_Runner");
use("Client_ServerProxy");

Class("Client_Application");
Class_Singleton();

Client_Application.simulationMode = false;

_.init = function () {
	this.runner = Client_Runner.get();
	this.server = new Client_ServerProxy;
	this.runRequest();
}

// events ---------------------------------------------------------------------

_.fireEvent_Command = function (domEvent) {
	var source = domEvent.target;
	if (source.tagName == 'menuitem') {
		var realSource = source.parentNode.parentNode;
		if (realSource.tagName == 'menu') {
			this.fireEvent('Click', domEvent, {});
		} else {
			var selectedIndex;
			if (realSource.tagName == 'button') {
				var children = source.parentNode.childNodes;
				selectedIndex = children.length;
				while (selectedIndex--) if (children[selectedIndex] == source) break;
			} else { // a menulist
				selectedIndex = realSource.selectedIndex;
			}
			this.fireEvent(
				'Select',
				{'target': realSource},
				{'selectedIndex': selectedIndex}
			);
		}
	} else {
		this.fireEvent('Click', domEvent, {});
	}
}

_.fireEvent_Select = function (domEvent) {
	var source = domEvent.target;
	var selectedIndex = source.selectedIndex;
	if (selectedIndex == -1) return; // listbox: mozilla fires strange events
	this.fireEvent
		('Select', {'target': source}, {'selectedIndex': selectedIndex });
}

_.fireEvent_Pick = function (domEvent) {
	var source = window.document.getElementById(domEvent.targetId);
	this.fireEvent('Pick', {'target': source}, {'color': source.color });
}

_.fireEvent_Change = function (domEvent)
	{ this.fireEvent('Change', domEvent, {'value': domEvent.target.value}) }

// private --------------------------------------------------------------------

_.fireEvent = function (name, domEvent, params) {
	var source   = domEvent.target;
	var sourceId = source.id;
	if (!sourceId) return; // event could come from some unknown place
	var event = {
		'source' : sourceId,
		'name'   : name,
		'checked': source.getAttribute('checked')
	};
	var key; for (key in params) event[key] = params[key];
	this.runRequest(event);
}

_.runRequest = function (event) {
	if (Client_Application.simulationMode) return;
	window.status = "Loading UI...";
	var response  = event? this.server.event(event): this.server.boot();
	window.status = "Running UI...";
	this.runner.run(response);
	window.status = "Done.";
}




